import enum
import tempfile
from typing import List, Optional, Union, Sequence

from ..model import BaseModel

VERSION = "2.1.0"

EventListen = enum.Enum(
    'listen',
    [('test', "test"), ('prerequest', "prerequest")],
)

Mode = enum.Enum('mode', [
    ('raw', 'raw'),
    ('urlencoded', 'urlencoded'),
    ('formdata', 'formdata'),
    ('file', 'file'),
    ('graphql', 'graphql'),
])


class Url(BaseModel):
    path: Sequence[str]
    host: Sequence[str]
    query: list
    variable: list


class OriginalRequest(BaseModel):
    method: str
    url: Url
    body: dict


class Header(BaseModel):
    key: str
    value: str


class Parameter(BaseModel):
    key: str
    value: str
    disabled: bool = False
    description: Optional[Union[str, dict]]


class Response(BaseModel):
    id: str
    name: str
    originalRequest: OriginalRequest
    status: str
    code: int
    header: Sequence[Header]
    body: str
    cookie: list
    _postman_previewlanguage: Optional[str]


class Auth(BaseModel):
    type: str


class QueryParam(Parameter):
    pass


class RequestBody(BaseModel):
    mode: Mode
    raw: str
    urlencoded: Optional[list]
    formdata: Optional[list]
    file: Optional[dict]
    graphql: Optional[dict]
    diabled: Optional[bool] = False


class Request(BaseModel):
    name: str
    description: dict
    method: str
    url: Url
    auth: Auth
    header: list
    body: Optional[RequestBody]


class Script(BaseModel):
    id: Optional[str]
    type: Optional[str]
    exec: Union[str, list]
    src: Optional[Union[str, Url]]
    name: Optional[str]

    def __add__(self, x):
        n = self.copy(deep=True)
        if isinstance(n.exec, str):
            n.exec = [n.exec]
        if isinstance(x.exec, str):
            x.exec = [x.exec]
        n.exec.extend(x.exec)
        return n


class Event(BaseModel):
    id: Optional[str]
    listen: str
    disabled: Optional[bool]
    script: Script


class Item(BaseModel):
    id: str
    name: str
    request: Request
    response: Sequence[Response]
    event: List[Event]


class Variable(BaseModel):
    id: str
    type: str
    value: Optional[str]


class Folder(BaseModel):
    id: str
    name: str
    item: Union[Sequence[Item], Item]
    event: list
    variable: Optional[Sequence[Variable]]


class Description(BaseModel):
    content: str
    type: str


class Info(BaseModel):
    name: str
    postman_id: str
    schema_: str
    description: Description

    class Config:
        fields = {'postman_id': '_postman_id', 'schema_': 'schema'}


class Collection(BaseModel):
    item: Sequence[Union[Item, Folder]]
    event: Optional[list]
    variable: Optional[Sequence[Variable]]
    info: Info

    def run(self, **kwargs):
        from ..postman_test import run_newman
        t = tempfile.mktemp()
        self.to_file(t)
        return run_newman(t, **kwargs)


class SaveDBResult(BaseModel):
    id: str
    object: Collection



import enum
from typing import Any, Dict, Optional, Union
from collections import defaultdict
from typing import List

from . import openapi_v3, postman_collection_v2
from ..model import BaseModel


class TestData(BaseModel):
    route: str
    in_: openapi_v3.ParameterIn
    key: str
    value: Any


class PostmanTest(enum.Enum):
    schema_validation = 'schema_validation'
    code = 'code'
    content_type = 'content_type'
    response_time = 'response_time'


class Expect(BaseModel):
    code: Optional[Union[int, str]]
    enabled: bool = True
    response_time: int = 200
    headers: Optional[Dict[str, str]]
    fake_data: bool = True
    enabled_tests: List[PostmanTest] = [
        PostmanTest.schema_validation,
        PostmanTest.code,
        PostmanTest.content_type,
        PostmanTest.response_time,
    ]


class TestConfig(BaseModel):
    variables: Optional[Union[Dict[str, str], str]]
    expect: Optional[Expect]
    make_global: Optional[Dict[str, str]]
    test: Optional[List[str]]
    prerequest: Optional[List[str]]
    headers: Optional[Dict[str, str]]


def list_of_test_data_to_params(
    route, test_data: List[TestData]
) -> openapi_v3.ParameterRequestData:
    d = defaultdict(lambda: {})
    for t in test_data:
        if t.route.lower() == route.lower():
            d[t.in_.value][t.key] = t.value
    return openapi_v3.ParameterRequestData(**dict(d))


class NewmanResult(BaseModel):
    stderr: Optional[str]
    stdout: str
    json_: Optional[dict]
    code: int

    class Config:
        fields = {'json_': 'json'}


class TestResult(BaseModel):
    result: NewmanResult
    code: int
    message: str


# class TestRun(BaseModel):
#     schema_: openapi_v3.OpenAPISchemaV3
#     test_data: List[TestData]
#     url: AnyHttpUrl
#
#     class Config:
#         fields = {'schema_': 'schema'}
#
#     def run(self, **kwargs) -> TestResult:
#         from ..util import run_newman
#         tmp_file = tempfile.mkfile()
#         self.schema_.to_file(tmp_file)
#         run_newman(tmp_file, **kwargs)
#         return TestResult(
#             result={'todo': 1},
#             code=1,
#             message="not implemented"
#         )


class OpenAPISchemaToPostmanRequest(BaseModel):
    schema_: Optional[openapi_v3.OpenAPISchemaV3]
    path: Optional[str] = None
    host: Optional[str]
    test_config: Optional[Dict[str, TestConfig]] = None
    test_data_file: Optional[str] = None
    test_data: Optional[List[TestData]] = None
    test_scripts: Optional[Dict[str, postman_collection_v2.Script]] = None
    prerequest_scripts: Optional[Dict[str, postman_collection_v2.Script]] = None  # noqa: E501
    collection_test_scripts: Optional[List[postman_collection_v2.Script]] = None  # noqa: E501
    collection_prerequest_scripts: Optional[List[postman_collection_v2.Script]] = None  # noqa: E501
    collection_global_variables: Optional[List[postman_collection_v2.Variable]] = None  # noqa: E501
    expect: Optional[Dict[str, Expect]] = None
    verbose: Optional[bool] = False

    class Config:
        fields = {'schema_': 'schema'}



import json

from shlex import quote
import subprocess
import tempfile

from .schema import NewmanResult
from .schema.postman_collection_v2 import (
    Event, Script,
)
from .util import (
    fingerprint, load_file, trim, uuidgen,
)


def run_newman(collection_file, host=None, verbose=None, json=False):
    cmdargs = []
    json_outfile, json_content = None, None
    if host:
        cmdargs.extend(['--env-var', f'baseUrl={quote(host)}'])
    if verbose:
        cmdargs.append('--verbose')
    cmdargs.extend(['--reporters', f'cli{",json" if json else ""}'])
    if json:
        json_outfile = tempfile.mktemp()
        cmdargs.extend(["--reporter-json-export", json_outfile])
    run_newman_args = ['newman', 'run', quote(collection_file), *cmdargs]
    e = subprocess.run(
        args=run_newman_args,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    print('EXEC', *run_newman_args)
    print('STDOUT', e.stdout.decode('utf-8'))
    print('STDERR', e.stderr.decode('utf-8'))
    if json_outfile:
        json_content = load_file(json_outfile, content_type='json')

    return NewmanResult(
        stderr=e.stderr.decode('utf-8'),
        stdout=e.stdout.decode('utf-8'),
        json_=json_content,
        code=e.returncode,
    )


def new_event(listen, script):
    if isinstance(script, list):
        if not len(script):
            return
        _script = script[0]
        for i in script[1:]:
            _script += i
        script = _script
    return Event(id=uuidgen(), listen=listen, script=script, disabled=False)


def javascript(name, exec):
    return Script(
        id=fingerprint([name, exec]),
        name=name,
        exec=trim(exec),
        type='text/javascript',
    )


def js_test_code(route, code):
    return javascript(
        name=f"{route} Test Code is {code}",
        exec="""
                pm.test("Status code is {code}", function () {{
                    return pm.response.to.have.status({code});
                }});
            """.format(code=code),
    )


def js_test_content_type(route, mimetype):
    return javascript(
        name=f"{route} Mimetype is {mimetype}",
        exec="""
                pm.test("Content-Type Header is {mimetype}", function () {{
                    return pm.expect(postman.getResponseHeader("Content-type")).to.be.eql("{mimetype}");
                }});
            """.format(mimetype=mimetype),
    )


def js_test_response_time(route, max_time_ms):
    return javascript(
        name=f"{route} responds in less than {max_time_ms}ms",
        exec="""
                pm.test("Response time is less than {max_time_ms}ms", function () {{
                    return pm.expect(pm.response.responseTime).to.be.below({max_time_ms});
                }});
            """.format(max_time_ms=max_time_ms),
    )


def js_test_validate_schema(route, schema, reference_schema):
    return javascript(
        name=f"{route} responds according to schema",
        exec="""
            pm.test('Schema is valid', function() {{
                var schema = {schema};
                tv4.addSchema({reference_schema});
                pm.expect(tv4.validate(pm.response.json(), schema)).to.be.true;
            }});
        """.format(schema=json.dumps(schema), reference_schema=json.dumps(reference_schema))
    )



import json

from pydantic import BaseModel as PyDanticBaseModel

from .util import DB, strip_nulls
from . import redis_handle


class BaseModel(PyDanticBaseModel):
    def to_file(self, path):
        with open(path, 'w') as f:
            json.dump(json.loads(self.json()), f)

    def save(self):
        dbo = DB(redis_handle(), model=self)
        dbo.save()
        return dbo

    def to_dict(self, no_empty=True):
        """Fixes serialization of dict()"""
        d = json.loads(self.json(by_alias=True))
        if no_empty:
            d = strip_nulls(d)
        return d

    def resolve_ref(self, schema):
        from .convert import OpenAPIToPostman
        if issubclass(schema.__class__, BaseModel):
            schema = schema.to_dict()
        return OpenAPIToPostman.find_ref(self.ref, schema)

    def get_safe(self, v):
        try:
            return self.__getattribute__(v)
        except AttributeError:
            pass
            # if default:
            #     return default

    class Config:
        extra = 'allow'



import enum

from typing import Any, Dict, List, Optional, Union, Set

from ..model import BaseModel

VERSION = "3.0.2"

HeaderType = lambda: Optional[Dict[str, Union[Header, Reference]]]
ExampleType = lambda: Optional[Dict[str, Union[Any, Reference]]]
ContentType = lambda: Optional[Dict[str, MediaType]]  # str is mimetype
ResponsesType = lambda: Dict[str, Union[Response, Reference]]
Style = Optional[str]
CallBackType = Dict[str, Dict[str, Dict]]


class ParameterRequestData(BaseModel):
    body: dict = dict()
    requestBody: dict = dict()
    path: dict = dict()
    query: dict = dict()
    header: dict = dict()
    cookie: dict = dict()


class ParameterIn(enum.Enum):
    body = 'body'
    request_body = 'requestBody'
    path = 'path'
    query = 'query'
    header = 'header'
    cookie = 'cookie'


class Reference(BaseModel):
    ref: str

    class Config:
        fields = {'ref': '$ref'}


class ExternalDocs(BaseModel):
    description: Optional[str]
    url: str


class ParameterSchema(BaseModel):
    type_: Optional[str]
    format: Optional[str]
    title: Optional[str]
    maximum: Optional[int]
    minimum: Optional[int]
    nullable: Optional[bool]

    class Config:
        fields = {'type_': 'type'}


class ServerVariable(BaseModel):
    default: str
    enum: Optional[List[str]]
    description: Optional[str]


class Server(BaseModel):
    url: str
    description: Optional[str]
    variables: Optional[Dict[str, ServerVariable]]


class Contact(BaseModel):
    name: Optional[str]
    url: Optional[str]
    email: Optional[str]


class License(BaseModel):
    name: str
    url: Optional[str]


class Info(BaseModel):
    description: Optional[str]
    termsOfService: Optional[str]
    title: Optional[str]
    version: Optional[str]
    contact: Optional[Contact]
    license: Optional[License]


class Properties(BaseModel):
    description: Optional[str]
    type: Optional[str]
    format: Optional[str]
    default: Optional[Any]
    example: Optional[str]
    title: Optional[str]
    multipleOf: Optional[int]
    maximum: Optional[int]
    exclusiveMaximum: Optional[bool]
    minimum: Optional[int]
    exclusiveMinimum: Optional[bool]
    maxLength: Optional[int]
    minLength: Optional[int]
    pattern: Optional[str]
    additionalItems: Optional[Union[bool, dict, Reference]]
    items: Optional[Union[dict, Reference]]
    maxItems: Optional[int]
    minItems: Optional[int]
    uniqueItems: Optional[bool]
    maxProperties: Optional[int]
    minProperties: Optional[int]
    required: Optional[Set[str]]
    properties: Optional[Union[dict, Reference]]
    additionalProperties: Optional[Union[bool, dict, Reference]]
    enum: Optional[Set[Any]]
    allOf: Optional[List[Union[dict, Reference]]]
    oneOf: Optional[List[Union[dict, Reference]]]
    anyOf: Optional[List[Union[dict, Reference]]]
    not_: Optional[Union[dict, Reference]]
    nullable: Optional[bool]
    discriminator: Optional[dict]
    readOnly: Optional[bool]
    writeOnly: Optional[bool]
    xml: Optional[dict]
    externalDocs: Optional[ExternalDocs]
    example: Optional[Any]
    deprecated: Optional[bool]

    class Config:
        fields = {'not_': 'not'}


class Encoding(BaseModel):
    contentType: Optional[str]
    headers: Optional[Dict[str, Union[Reference, dict]]]
    style: Style
    explode: Optional[bool]
    allowReserved: Optional[bool]


class SchemaObject(BaseModel):
    description: Optional[str]
    required: Optional[List[str]]
    example: Optional[Any]
    examples: ExampleType()
    encoding: Optional[Dict[str, Encoding]]
    type: Optional[str]
    format: Optional[str]
    properties: Optional[Dict[str, Properties]]
    allOf: Optional[List[Union[dict, Reference]]]

    class Config:
        fields = {'type_': 'type'}


class MediaType(BaseModel):
    schema_: Union[Reference, SchemaObject, None]

    class Config:
        fields = {'schema_': 'schema'}


class Header(BaseModel):
    description: Optional[str]
    required: Optional[bool]
    schema_: ParameterSchema
    deprecated: Optional[bool]
    allowEmptyValue: Optional[bool]
    style: Style
    explode: Optional[bool]
    allowReserved: Optional[bool]
    example: Optional[Any]
    examples: ExampleType()
    content: ContentType()

    class Config:
        fields = {'schema_': 'schema'}


class Response(BaseModel):
    description: str
    content: ContentType()
    headers: HeaderType()


class Parameter(BaseModel):
    name: str
    in_: ParameterIn
    description: Optional[str]
    required: Optional[bool]
    schema_: ParameterSchema
    deprecated: Optional[bool]
    allowEmptyValue: Optional[bool]
    style: Style
    explode: Optional[bool]
    allowReserved: Optional[bool]
    example: Optional[Any]
    examples: ExampleType()
    content: ContentType()

    class Config:
        fields = {'schema_': 'schema', 'in_': 'in'}


class RequestBody(BaseModel):
    description: Optional[str]
    content: Dict[str, MediaType]
    required: Optional[bool]


class Operation(BaseModel):
    operationId: Optional[str]
    summary: Optional[str]
    responses: ResponsesType()
    parameters: Optional[List[Union[Parameter, Reference]]]
    tags: Optional[List[str]]
    externalDocs: Optional[ExternalDocs]
    requestBody: Optional[Union[RequestBody, Reference]]
    callbacks: Optional[Dict[str, Union[CallBackType, Reference]]]
    deprecated: Optional[bool]
    security: Optional[List[Dict[str, List]]]
    servers: Optional[List[Server]]


# class Components(BaseModel):
#     schemas: Optional[Dict[str, Union[SchemaObject, Reference]]]
#     responses: Optional[Dict[str, ResponsesType()]]
#     parameters: Optional[Dict[str, Union[Parameter, Reference]]]


class Path(BaseModel):
    # summary: Optional[str]
    # description: Optional[str]
    get: Optional[Operation]
    put: Optional[Operation]
    post: Optional[Operation]
    delete: Optional[Operation]
    options: Optional[Operation]
    head: Optional[Operation]
    patch: Optional[Operation]
    trace: Optional[Operation]
    # servers: Optional[List[Server]]
    # parameters: Optional[List[Union[Parameter, Reference]]]


class OpenAPISchemaV3(BaseModel):
    openapi: str
    paths: Dict[str, Path]
    info: Optional[Info]
    servers: Optional[List[Server]]
    components: Optional[dict]
    security: Optional[List[Dict[str, List]]]
    externalDocs: Optional[ExternalDocs]


class SaveDBResult(BaseModel):
    id: str
    object: OpenAPISchemaV3



from collections import ChainMap
from typing import Any, Callable, Iterable, Union
import functools
import importlib
import hashlib
import jinja2
import json
import os
import random
import re
import secrets
import string
from types import GeneratorType
import yaml

import hypothesis
from hypothesis import given
from hypothesis_jsonschema._from_schema import from_schema

from . import logger, redis_handle, Settings

RE_WORDCHARS = re.compile('^\w+$')  # noqa: W605

HTTP_VERBS = [
    "get",
    "put",
    "post",
    "delete",
    "options",
    "head",
    "patch",
    "trace",
]

FILTERS = {
    str: lambda x: len(x) >= Settings().test_data_str_min_length,
    int: lambda x: x >= Settings().test_data_int_min,
    'words': lambda x: RE_WORDCHARS.match(x),
}

# def filter_min_length(x):
#     min_len = Settings().test_data_str_min_length
#     if isinstance(x, str):
#         logger.warning(f"STR {len(x)}")
#         return len(x) >= min_len
#     elif isinstance(x, int):
#         logger.warning(f"INT {x}")
#         return x >= min_len
#     else:
#         logger.warning(f"Don't know how to check length of type({type(x)})")
#         return True


def strip_nulls(obj):
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(strip_nulls(x) for x in obj if x is not None)
    elif isinstance(obj, dict):
        return type(obj)(
            (strip_nulls(k), strip_nulls(v))
            for k, v in obj.items()
            if k is not None and v is not None
        )
    else:
        return obj


def blind_load(content):
    load_map = {
        'json': json.loads,
        'yaml': yaml.safe_load,
    }
    if content.lstrip().startswith('{'):
        content_type = "json"
    else:
        content_type = "yaml"
    try:
        parsed_content = load_map[content_type](content)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError, json.decoder.JSONDecodeError) as e:
        logger.warn(e)
        map_type = "yaml" if content_type == "json" else "json"
        parsed_content = load_map[map_type](content)
    return parsed_content


def is_local_path(s):
    """Does not consider paths above cwd to be valid."""
    if (
        isinstance(s, str)
        and os.path.exists(s)
        and os.path.abspath(s).startswith(os.getcwd())
    ):
        return True


def load_file(f, content_type=None):
    if f.endswith('.yaml') or f.endswith('.yml') or content_type == 'yaml':
        load_f = yaml.safe_load
    elif f.endswith('.json') or content_type == 'json':
        load_f = json.load
    else:
        raise ValueError(f"Unknown file type: {f}")
    with open(f, 'r') as fp:
        return load_f(fp)


def trim(s):
    return s.rstrip().lstrip()


def cls_from_str(name):
    # Import the module .
    components = name.split('.')
    module = importlib.import_module(
        '.'.join(components[:len(components) - 1]),
    )
    a = module.__getattribute__(components[-1])
    return a


def fingerprint(payload: Any):
    if isinstance(payload, Iterable) and not isinstance(payload, dict):
        payload = [p for p in payload]
    payload = repr(payload)
    return hashlib.sha512(payload.encode('utf-8')).hexdigest()


class DB:
    def __init__(self, r, model=None, uid=None):
        if uid:
            model = self.load_model_from_uid(r, uid)
            if not model:
                raise ValueError("Unable to load by uid")
        elif model:
            self.uid = fingerprint(model.dict())
        else:
            raise ValueError("Model or uid is required")
        self.model = model
        self.cache_ttl = 60 * 60 * 24 * 7
        self.klass = type(self.model).__module__ + '.'\
            + type(self.model).__name__
        self.json = json.dumps({
            'data': self.model.to_dict(),
            'class': self.klass,
        })
        self.r = r

    @classmethod
    def _get(cls, r, uid):
        logger.debug(f'Getting {uid} from Redis.')
        return r.get(uid).decode('utf-8')

    @classmethod
    def load_model_from_uid(cls, r, uid):
        raw = cls._get(r, uid)
        if raw:
            raw = json.loads(raw)
            return cls_from_str(raw['class']).construct(**raw['data'])

    def save(self):
        logger.debug(f'Creating {repr(self)} in Redis.')
        return self.r.setex(
            self.uid, self.cache_ttl,
            self.json.encode('utf-8'),
        )


def save_db(o):
    odb = DB(redis_handle(), model=o)
    if not odb.save():
        raise ValueError(f"Unable to save {o}(uid: {odb.uid}) to the DB.")
    logger.debug(f"New {type(o)}: {odb.uid}")
    return o.construct(**{'id': odb.uid, 'object': o.to_dict()})


def load_db(id):
    return DB.load_model_from_uid(redis_handle(), id)


HTTP_REASON_CLASS = {
    1: "Informational - Request received, continuing process",
    2: "Success - The action was successfully received, understood, and accepted",  # noqa: E501
    3: "Redirection - Further action must be taken in order to complete the request",  # noqa: E501
    4: "Client Error - The request contains bad syntax or cannot be fulfilled",
    5: "Server Error - The server failed to fulfill an apparently valid request",  # noqa: E501
}


HTTP_REASONS = {
    200: "OK",
    407: "Proxy Authentication Required",
    408: "Request Time-out",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Request Entity Too Large",
    414: "Request-URI Too Large",
    415: "Unsupported Media Type",
    416: "Requested range not satisfiable",
    417: "Expectation Failed",
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Time-out",
    505: "HTTP Version not supported",
}


def get_http_reason(code):
    if (
        isinstance(code, str)
        and not code.isdigit() and code[0].isdigit()
        and len(code) == 3
    ):
        return HTTP_REASON_CLASS.get(int(code[0]))
    reason = HTTP_REASONS.get(int(code))
    if not reason:
        reason = HTTP_REASON_CLASS.get(int(str(code)[0]))
    return reason


def render_jinja2(template: str, **kwargs) -> str:
    return jinja2.Template(template).render(**kwargs)


def uuidgen(*_, **__):
    t = '-'.join([secrets.token_hex(i // 2) for i in [8, 4, 4, 4, 12]])
    return t


def gen_string(length, charset=string.printable, choice_f=random.choice):
    return ''.join([choice_f(charset) for i in range(0, length)])


def flatten_iterables_in_dict(d: dict, index=0, no_null=True, min_length=0):
    nd = {}
    for k, v in d.copy().items():
        if isinstance(v, dict):
            nd[k] = flatten_iterables_in_dict(
                v, index=index, no_null=no_null,
                min_length=min_length,
            )
        elif isinstance(v, Iterable):
            if not len(v):
                continue
            if (
                (no_null and v[index] is None)
                or (isinstance(v[index], str) and len(v[index]) <= min_length)
            ):
                for i in v:
                    if (not no_null or i is not None) and len(i) >= min_length:
                        nd[k] = i
                        break
            else:
                nd[k] = v[index]
        else:
            nd[k] = v
    return nd


class TemplateMap:
    def __init__(self, map: dict, defaults: dict, template_args: dict):
        self.map = ChainMap(map, defaults)
        self.res = self.parse_map(self.map, template_args)

    @classmethod
    def parse_map_item(
        cls, map: dict, k: Any,
        v: Union[Callable, str, dict, Iterable],
        template_args: dict,
    ):
        if isinstance(v, Callable):
            res = v(map, k, template_args)
            if isinstance(res, dict):
                return cls.parse_map(res, template_args)
            else:
                return cls.parse_map_item(map, k, res, template_args)
        elif isinstance(v, str):
            return render_jinja2(v, **template_args)
        elif isinstance(v, dict):
            return cls.parse_map(v, template_args)
        elif isinstance(v, Iterable):
            res = []
            for i in v:
                res.append(cls.parse_map_item(map, k, i, template_args))
            return res
        else:
            raise ValueError(f"Unhandled type: {type(v)}")

    @classmethod
    def parse_map(cls, map: dict, template_args: dict) -> dict:
        res = {}
        for k, v in map.items():
            res[k] = cls.parse_map_item(map, k, v, template_args)
        return res


def generate_from_schema(schema, no_empty=True, retry=5):
    test_data = []
    settings = Settings()
    if (
        schema.get('type') == 'string'
        and not schema.get('minLength')
        and not schema.get('maxLength')
        and not schema.get('format')
    ):
        schema['minLength'] = settings.test_data_str_min_length
        # schema['pattern'] = '^\w+$'
        generate_func = from_schema(schema)
        generate_func = generate_func.filter(FILTERS[str])
    elif schema.get('type') == 'integer' and not schema.get('minimum'):
        schema['minimum'] = settings.test_data_int_min
        generate_func = from_schema(schema)
        generate_func = generate_func.filter(FILTERS[int])
    else:
        generate_func = from_schema(schema)

    @given(generate_func)
    def f(x):
        if x or not no_empty:
            test_data.append(x)
    passed = False
    while retry > 0 or not passed:
        try:
            f()
            passed = True
            retry = 0
        except (hypothesis.errors.Unsatisfiable, hypothesis.errors.FailedHealthCheck):
            retry -= 1
    if not passed:
        raise hypothesis.errors.Unsatisfiable("Max retries hit")
    yield test_data


def pick_one(gen: GeneratorType, strategy="random"):
    """Given a generator which yields an iterable, get an element."""
    # it seems that the data from generate_from_schema
    # is better if you pick randomly
    # often enough the first element is rather boring like 0 or '0'
    if "rand" in strategy.lower():
        return random.choice(next(gen))
    else:
        return next(gen)[0]


def hashable_lru(func):
    cache = functools.lru_cache(maxsize=1024)

    def deserialise(value):
        try:
            return json.loads(value)
        except Exception:
            return value

    def func_with_serialized_params(*args, **kwargs):
        _args = tuple([deserialise(arg) for arg in args])
        _kwargs = {k: deserialise(v) for k, v in kwargs.items()}
        return func(*_args, **_kwargs)

    cached_function = cache(func_with_serialized_params)

    @functools.wraps(func)
    def lru_decorator(*args, **kwargs):
        _args = tuple([json.dumps(arg, sort_keys=True) if type(arg) in (list, dict) else arg for arg in args])  # noqa: E501
        _kwargs = {k: json.dumps(v, sort_keys=True) if type(v) in (list, dict) else v for k, v in kwargs.items()}  # noqa: E501
        return cached_function(*_args, **_kwargs)
    lru_decorator.cache_info = cached_function.cache_info
    lru_decorator.cache_clear = cached_function.cache_clear
    return lru_decorator



import logging

import redis
from pydantic import BaseSettings

FAKE_REDIS_SERVER = None

class Settings(BaseSettings):
    redis_host: str = 'localhost'
    redis_port: int = 6379
    testing: bool = False
    log_level: str = 'WARNING'
    test_data_str_min_length: int = 1
    test_data_int_min: int = 1

    class Config:
        env_file = '.env'
        fields = {
            'redis_dsn': {
                'env': 'redis_url',
            },
            'testing': {
                'env': 'testing'
            }
        }


def redis_handle():
    settings = Settings().dict()
    if settings['testing']:
        import fakeredis
        global FAKE_REDIS_SERVER
        if not FAKE_REDIS_SERVER:
            FAKE_REDIS_SERVER = fakeredis.FakeServer()
        r = fakeredis.FakeRedis(server=FAKE_REDIS_SERVER)
    else:
        r = redis.Redis(
            host=settings['redis_host'],
            port=settings['redis_port'],
        )
    return r


logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(Settings().log_level)


from . import cli  # noqa: E402, F401



import os
from collections import defaultdict, ChainMap, Counter, namedtuple
from typing import Generator, Iterable, List, Optional
import json
import re
import yaml

from .postman_test import (
    new_event, javascript, js_test_code,
    js_test_content_type, js_test_response_time,
    js_test_validate_schema,
)
from .schema import (
    Expect, OpenAPISchemaToPostmanRequest, PostmanTest,
    TestData, TestConfig,
    list_of_test_data_to_params,
)
from .schema import openapi_v3 as oapi
from .schema.openapi_v3 import (
    Operation, OpenAPISchemaV3, ParameterIn,
    ParameterRequestData, Reference, SchemaObject,
)
from .schema.postman_collection_v2 import (
    Auth, Collection, Item,
    Request, RequestBody, Response, OriginalRequest, Header,
    Info, Description, Parameter, Url, Variable,
)
from .util import (
    blind_load,
    fingerprint, flatten_iterables_in_dict, generate_from_schema,
    get_http_reason, hashable_lru, is_local_path, load_file, pick_one,
    uuidgen, trim, HTTP_VERBS,
)
from . import logger

RE_PATH_VARIABLE = re.compile('\{(.*?)\}')  # noqa: W605
RE_PATH_GLOBAL_VARIABLE = re.compile('\{\{(.*?)\}\}')  # noqa: W605
RE_PATH_VARIABLE_SEGMENT = re.compile('(\(\{(.*?)\}\))')  # noqa: W605
RE_PATH_CONVERTED_VARIABLE_SEGMENT = re.compile('(\{\{(.*?)\}\})')  # noqa: W605

Route = namedtuple('Route', ['verb', 'path', 'operation'])
ReferenceSearch = namedtuple('ReferenceSearch', ['ref', 'schema'])
PostmanConfig = namedtuple('PostmanConfig', [
    'expect',
    'test_data',
    'test_scripts',
    'prerequest_scripts',
    'collection_global_variables',
    'collection_test_scripts',
    'collection_prerequest_scripts',
])
PostmanVariables = namedtuple('PostmanVariables', [
    'global_',
    'query',
    'url',
    'header',
    'body',
])


class OpenAPIToPostman:
    def __init__(
        self,
        request: Optional[OpenAPISchemaToPostmanRequest] = None,
        **kwargs,
    ):
        if not request:
            request = OpenAPISchemaToPostmanRequest(**kwargs)
        self.schema = None
        self.host = request.host
        self.max_ref_depth = 5
        self.ref_depth = Counter()
        self.strict = True
        self.test_scripts = defaultdict(lambda: [], request.test_scripts or [])
        self.prerequest_scripts = defaultdict(lambda: [], request.prerequest_scripts or [])  # noqa: E501
        self.collection_test_scripts = request.collection_test_scripts or []
        self.collection_prerequest_scripts = request.collection_prerequest_scripts or []  # noqa: E501
        self.collection_global_variables = request.collection_global_variables or []
        self.default_expect = Expect()
        self.expect = request.expect or defaultdict(self.get_default_expect)
        path = request.path
        schema = request.schema_
        if path:
            if path.endswith('.json'):
                self.schema = OpenAPISchemaV3.parse_file(path)
            elif path.endswith('.yaml') or path.endswith('.yml'):
                with open(path, "r") as f:
                    self.schema = yaml.safe_load(f)
            else:
                raise ValueError(f"Unknown file type for: {path}")
        elif schema:
            self.schema = schema
        else:
            raise ValueError(
                "Either path to schema or schema is a required field",
            )
        if isinstance(self.schema, OpenAPISchemaV3):
            self.schema = self.schema.to_dict(no_empty=False)
        # Validate schema
        self.schema = OpenAPISchemaV3(**self.schema)
        self.test_data = self.load_test_data(
            request.test_data or [],
            request.test_data_file,
            request.test_config,
        )

    @classmethod
    def load_remote_refs(cls, schema_path):
        if isinstance(schema_path, str) and schema_path.startswith('http'):
            import requests
            ref_path = None
            if '#' in schema_path:
                ref_path = schema_path.split('#')[-1]
            try:
                logger.debug(f"Fetching remote reference: {schema_path}")
                schema_path = blind_load(
                    requests.get(schema_path).content.decode('utf-8'),
                )
                if ref_path:
                    schema_path = cls.find_ref(f'#/{ref_path}', schema_path)
            except Exception as e:
                logger.error(f"Exception: {type(e)}: {e}")
                raise e
        return schema_path

    @classmethod
    def load_local_refs(cls, schema_path):
        if isinstance(schema_path, str) and os.path.exists(schema_path):
            schema_path = load_file(schema_path)
        return schema_path

    @classmethod
    @hashable_lru
    def find_ref(cls, ref: str, schema_path):
        ref_path = ref.split('/')[1:]
        while len(ref_path):
            seek = ref_path.pop(0).replace('~1', '/').replace('~0', '~')
            if seek:
                if isinstance(schema_path, list) and seek.isdigit():
                    schema_path = schema_path[int(seek)]
                else:
                    schema_path = schema_path[seek]
        schema_path = cls.load_remote_refs(schema_path)
        schema_path = cls.load_local_refs(schema_path)

        return schema_path


    @classmethod
    def path_parts(cls, path: str) -> list:
        url = []
        for part in path.split('/')[1:]:
            is_path_variable = re.match('^{(\w+)}$', part)  # noqa: W605
            is_path_segment = RE_PATH_VARIABLE.findall(part)
            # queries = RE_PATH_VARIABLE_SEGMENT.findall(part)
            if is_path_variable:
                part = f':{is_path_variable.group(1)}'
            elif is_path_segment:
                for group in is_path_segment:
                    part = part.replace(f"{{{group}}}", f"{{{{{group}}}}}")
            if '(' in part and ')' in part:
                part = part.replace('(', '').replace(')', '')
            url.append(part)
        return url

    def get_default_expect(self):
        return self.default_expect

    def test_config_to_postman_config(self, test_config) -> PostmanConfig:
        test_data = []
        expect = defaultdict(self.get_default_expect)
        test_scripts = defaultdict(lambda: [])
        prerequest_scripts = defaultdict(lambda: [])
        collection_global_variables = []
        collection_test_scripts = []
        collection_prerequest_scripts = []
        for route, td_item in test_config.items():
            i = TestConfig(**td_item)
            if isinstance(i.variables, str):
                variables = load_file(i.variables)
            else:
                variables = i.variables
            if route.lower() == 'collection':
                collection_global_variables.extend([
                    Variable(id=k, type='string', value=v)
                    for k, v in (variables or {}).items()
                ])
                if i.expect:
                    self.default_expect = i.expect
                if i.test:
                    collection_test_scripts.extend([
                        javascript(
                            exec=t,
                            name=f"{route} cmd({fingerprint(t)})"
                        ) for t in i.test
                    ])
                if i.prerequest:
                    collection_prerequest_scripts.extend([
                        javascript(
                            exec=t,
                            name=f"{route} cmd({fingerprint(t)})"
                        ) for t in i.prerequest
                    ])
                continue
            verb, path = route.split(' ')
            route = f'{verb.upper()} {path}'
            for variable, response_path in (i.make_global or {}).items():
                response_path = ''.join([
                    '["' + trim(p) + '"]'
                    for p in response_path.split('.')[1:]
                ])
                debug_global = lambda x: javascript(
                    name=f'[{x}] Debug {variable}',
                    exec=f'console.log("[{x}] GLOBAL({variable}):", pm.globals.get("{variable}"));',  # noqa: E501
                )
                collection_prerequest_scripts.append(
                    debug_global('collection_prerequest'),
                )
                prerequest_scripts[route].append(
                    debug_global('prerequest'),
                )
                test_scripts[route].append(javascript(
                    name=f'Set response of {route}: JSON_RESPONSE{response_path} to {variable}',  # noqa: E501
                    exec="""
                        pm.test('set {variable}', function() {{
                            pm.globals.set("{variable}", pm.response.json(){response_path});
                        }});
                    """.format(variable=variable, response_path=response_path),  # noqa: E501
                ))
            # self.collection_global_variables.append(
            # Variable(id=variable, type='string', value='default'))
            for k, v in (variables or {}).items():
                if is_local_path(v):
                    v = load_file(v)
                test_data.extend([
                    TestData(route=route, in_=in_, key=k, value=v)
                    for in_ in list(ParameterIn)
                ])
            if i.expect:
                expect[route] = i.expect

            if i.test:
                test_scripts[route].extend([
                    javascript(
                        exec=t,
                        name=f"{route} cmd({fingerprint(t)})",
                    ) for t in i.test
                ])
            if i.prerequest:
                prerequest_scripts[route].extend([
                    javascript(
                        exec=t,
                        name=f"{route} cmd({fingerprint(t)})",
                    ) for t in i.prerequest
                ])
        for verb, path, operation in self.routes:
            route_str = f"{verb.upper()} {path}"
            for code, response in operation.responses.items():
                if code == 'default':
                    code = 500
                appended_test_scripts = False
                for mimetype, media_type in (response.content or {}).items():
                    if appended_test_scripts:
                        continue
                    candidate_route = (
                        (expect[route_str].code == code and expect[route_str].enabled)
                        or str(code).startswith('2')
                    )
                    if candidate_route:
                        enabled_tests = expect[route_str].enabled_tests
                        if mimetype == 'application/json':
                            schema_ = self._resolve_object(media_type.schema_ or {})
                            mt_props = self._resolve_object(
                                schema_.get('properties') or {},
                            )
                            if PostmanTest.schema_validation in enabled_tests:
                                test_scripts[route_str].append(
                                    js_test_validate_schema(route_str, mt_props, self.schema.to_dict())
                                )
                        if PostmanTest.code in enabled_tests:
                            test_scripts[route_str].append(
                                js_test_code(route_str, code),
                            )
                        if PostmanTest.content_type in enabled_tests:
                            test_scripts[route_str].append(
                                js_test_content_type(route_str, mimetype),
                            )
                        if PostmanTest.response_time in enabled_tests:
                            test_scripts[route_str].append(
                                js_test_response_time(
                                    route_str,
                                    expect[route_str].response_time,
                                )
                            )
                        appended_test_scripts = True
        return PostmanConfig(
            expect, test_data, test_scripts,
            prerequest_scripts, collection_global_variables,
            collection_test_scripts, collection_prerequest_scripts,
        )

    def load_test_data(
        self, test_data, test_data_file=None,
        test_config=None,
    ):
        if not (test_data_file or test_config):
            return test_data
        if test_data_file:
            test_config = load_file(test_data_file)
        postman_config = self.test_config_to_postman_config(test_config)
        self.test_scripts.update(postman_config.test_scripts)
        self.prerequest_scripts.update(postman_config.prerequest_scripts)
        self.collection_global_variables.extend(
            postman_config.collection_global_variables,
        )
        self.collection_test_scripts.extend(
            postman_config.collection_test_scripts,
        )
        self.collection_prerequest_scripts.extend(
            postman_config.collection_prerequest_scripts,
        )
        test_data.extend(postman_config.test_data)
        return test_data

    @property
    def paths(self):
        for path in self.schema.paths:
            yield (path, self.schema.paths[path])

    @classmethod
    def guess_resource(cls, path: str):
        parts = cls.path_parts(path)
        last_part = None
        for i, part in enumerate(parts):
            is_variable = RE_PATH_GLOBAL_VARIABLE.match(part)\
                or part.startswith(':')
            segments = RE_PATH_CONVERTED_VARIABLE_SEGMENT.findall(part)
            for segment, var in segments:
                part = part.replace(segment, '')
            # If this segment is a variable, return the last one
            if is_variable and len(parts) - i == 1:
                return last_part
            last_part = part
        return last_part

    @classmethod
    def order_routes_by_resource(
        cls, routes: Iterable[Route],
        verb_ordering: List = list(['post', 'put', 'get', 'patch', 'delete'])
    ) -> List[Route]:
        """Identify REST resources and order CRUD operations safely."""
        by_resource = defaultdict(lambda: {})
        output = []
        for route in routes:
            verb, path, operation = route
            by_resource[cls.guess_resource(path)][verb] = route
        for resource, routes_by_verb in by_resource.items():
            for verb in verb_ordering:
                if routes_by_verb.get(verb):
                    output.append(routes_by_verb[verb])
        return output

    def fake_data_from_route_schema(
        self, verb: str, path: str, operation: Operation,
    ) -> ParameterRequestData:
        if not self.expect[f'{verb.upper()} {path}'].fake_data:
            logger.debug(f"Skipping fake_data generation for {verb} {path}")
            return ParameterRequestData()
        d = defaultdict(lambda: {})
        all_path_vars = RE_PATH_VARIABLE.findall(path)
        parameters = self._resolve_object(operation.parameters)

        if parameters:
            for parameter in operation.parameters:
                parameter = self._resolve_object(
                    parameter,
                    new_cls=oapi.Parameter,
                )
                in_ = parameter.in_.value
                param_schema = parameter.schema_
                if in_ == 'body':
                    properties = param_schema.properties or {}
                    if isinstance(properties, Reference):
                        properties = properties.resolve_ref(self.schema)
                    for param, param_schema in properties.items():
                        d[in_][param] = pick_one(
                            generate_from_schema(param_schema),
                        )
                else:
                    d[in_][parameter.name] = pick_one(
                        generate_from_schema(param_schema.to_dict()),
                    )
        request_body = self._resolve_object(
            operation.requestBody,
            new_cls=oapi.RequestBody,
        )
        if request_body:
            for mimetype, media_type in request_body.content.items():
                schema_ = self._resolve_object(media_type.schema_ or {})
                mt_props = self._resolve_object(
                    schema_.get('properties') or {},
                )
                for name, prop in mt_props.items():
                    prop = self._resolve_object(prop)
                    if not isinstance(prop, dict):
                        prop = prop.to_dict()
                    if prop.get('ref') or prop.get('$ref'):
                        logger.error(f"Unresolved reference in media type! {media_type}")  # noqa; E501
                    d['requestBody'][name] = pick_one(
                        generate_from_schema(prop),
                    )
        for path_var in set(all_path_vars).difference(set(d.get('path', []))):
            logger.warning(
                f"Path variable {path_var} isn't defined as a parameter "
                f"generating test data for it assuming it's a string."
            )
            d['path'][path_var] = pick_one(
                generate_from_schema({'type': 'string'}),
            )
        d = dict(d)
        return ParameterRequestData(**d)

    @property
    def routes(self) -> Generator[Route, None, None]:
        for path, operation in self.paths:
            for verb, operation in [
                (verb, operation.get_safe(verb))
                for verb in HTTP_VERBS if operation.get_safe(verb)
            ]:
                yield Route(verb, path, operation)

    @property
    def info(self):
        return self.schema.info

    @property
    def title_version(self):
        return f"{self.info.title} {self.info.version}"

    def operation_param_examples(self, operation: Operation):
        examples = defaultdict(lambda: defaultdict(lambda: []))
        for param in self._resolve_object(operation.parameters or []):
            param = self._resolve_object(param, new_cls=oapi.Parameter)
            param_in = param.in_.value
            if param_in == 'body':
                examples['body'] = []
                target = examples['body']
            else:
                target = examples[param_in][param.name]
            if param.example:
                target.extend([param.example])
            if param.examples:
                target.extend(param.examples)
        return examples

    def _resolve_object(self, o, new_cls=None, deep=False):
        if isinstance(o, Reference):
            logger.debug(f"Resolving reference {o.ref} Try #{self.ref_depth[o.ref]}")  # noqa; E501
            if self.ref_depth[o.ref] > self.max_ref_depth:
                if self.strict:
                    raise ValueError(f"Max reference recursion reached for {o.ref}")  # noqa; E501
                else:
                    return
            self.ref_depth[o.ref] += 1
            o = o.resolve_ref(self.schema)
        elif '$ref' in dir(o):
            if isinstance(o, SchemaObject):
                logger.warning("SchemaObject should be a Reference object, catching serialization issue...")
                o = self._resolve_object(
                    Reference(
                        **{'$ref': o.__getattribute__('$ref')},
                    ),
                    new_cls=new_cls,
                    deep=deep,
                )
            else:
                logger.error(
                    "Possible serialization issue with object,"
                    f" object of type ({type(o)}) shouldn't be a reference: ({o})"
                )
        elif isinstance(o, dict) and o.get('$ref'):
            o = self._resolve_object(
                Reference(**o),
                new_cls=new_cls,
                deep=deep,
            )
        if 'to_dict' in dir(o):
            o = o.to_dict()
        if deep and isinstance(o, dict):
            for k, v in o.items():
                if (isinstance(v, dict) and v.get('$ref')) or isinstance(v, Reference):
                    if isinstance(v, Reference):
                        v = v.to_dict()
                    o[k] = self._resolve_object(Reference(**v), deep=True)
                else:
                    o[k] = self._resolve_object(v, deep=True)
        if isinstance(o, dict) and new_cls:
            o = new_cls(**o)
        return o

    def convert_parameters(self, verb, path, operation: Operation):
        return ParameterBuilder(
            self, verb, path, operation, self.test_data,
        ).build()

    def _generate_postman_collections(self):
        build_url = lambda path, vars=None, query=[]: Url(
            host=["{{baseUrl}}"],
            path=self.path_parts(path),
            query=query,
            variable=vars or [],
        )
        items = []
        global_variables = self.collection_global_variables or []
        ordered_routes = self.order_routes_by_resource(self.routes)
        for verb, path, operation in ordered_routes:
            responses = []
            route_str = f"{verb.upper()} {path}"
            for code, response in operation.responses.items():
                if code == 'default':
                    code = 500
                if isinstance(code, str) and 'x' in code.lower():
                    code = code.lower().replace('x', '0')
                http_reason = get_http_reason(code)
                response = self._resolve_object(
                    response, new_cls=oapi.Response,
                )
                for mimetype, route_definition in (
                    response.content or {'text/html': {}}
                ).items():
                    responses.append(Response(
                        id=uuidgen(),
                        name=response.description,
                        originalRequest=OriginalRequest(
                            url=build_url(path),
                            method=verb.upper(),
                            body={},
                        ),
                        code=int(code),
                        status=http_reason,
                        header=[
                            Header(key='Content-Type', value=mimetype),
                        ],
                        cookie=[],
                        body=response.description,
                    ))
            (
                new_globals, query, request_url_variables,
                request_header, request_body
            ) = self.convert_parameters(verb, path, operation)
            global_variables.extend(new_globals)
            logger.debug('GLOBALS: ' + repr(new_globals))
            logger.debug('REQUEST: ' + repr(request_url_variables))
            items.append(
                Item(
                    id=uuidgen(),
                    name=operation.summary or route_str,
                    request=Request(
                        auth=Auth(type='noauth'),
                        url=build_url(
                            path, vars=request_url_variables, query=query,
                        ),
                        method=verb.upper(),
                        name=operation.summary or route_str,
                        description={},
                        body=request_body,
                        header=request_header,
                    ),
                    response=responses,
                    event=[
                        e for e in [
                            new_event(
                                'test',
                                self.test_scripts.get(route_str, []),
                            ),
                            new_event(
                                'prerequest',
                                self.prerequest_scripts.get(route_str, []),
                            ),
                        ] if e
                    ],
                )
            )

        return [
            global_variables, items
        ]

    def to_postman_collection_v2(self):
        global_variables, items = self._generate_postman_collections()
        return Collection(
            event=[
                e for e in [
                    new_event('test', self.collection_test_scripts),
                    new_event(
                        'prerequest',
                        self.collection_prerequest_scripts,
                    ),
                ] if e
            ],
            variable=[
                Variable(id='baseUrl', type='string', value=self.host or '/'),
                *global_variables,
            ],
            item=items,
            info=Info(
                _postman_id=uuidgen(),
                name=self.info.title,
                schema='https://schema.getpostman.com/json/collection/v2.1.0/collection.json',  # noqa: E501
                description=Description(
                    content=self.info.get_safe('description')\
                        or self.title_version,
                    type='text/plain',
                ),
            )
        )


class ParameterBuilder:
    def __init__(self, mormo, verb, path, operation, test_data):
        self.mormo, self.verb, self.path, self.operation, self.test_data = (
            mormo, verb, path, operation, test_data,
        )
        self.params = mormo._resolve_object(operation.parameters or [])
        self.config_test_data = list_of_test_data_to_params(
            f"{verb} {path}",
            self.test_data,
        ).dict()
        self.fake_data = mormo.fake_data_from_route_schema(
            verb, path, operation,
        ).dict()
        self.examples = flatten_iterables_in_dict(
            mormo.operation_param_examples(operation),
        )
        self.path_vars = [p[1:] for p in mormo.path_parts(path) if p.startswith(':')]

    def get_mapped_value(self, v):
        return ChainMap(
            self.config_test_data.get(v, {}),
            self.examples.get(v, {}),
            self.fake_data.get(v, {}),
        )

    def build_test_data_from_param(self, param, mapped_value):
        return Parameter(
            key=param.name,
            value=str(mapped_value[param.name]),
        )

    def get_path_param_variables(self, url, param):
        global_ = []
        segment_vars = set(RE_PATH_VARIABLE.findall(self.path))\
            .difference(self.path_vars)
        found_var_location = False
        mapped_value = self.get_mapped_value('path')
        for v in segment_vars:
            if v == param.name:
                if v in mapped_value:
                    global_.append(
                        Variable(
                            id=v, type='string',
                            value=str(mapped_value[v])
                        )
                    )
                    found_var_location = True
                else:
                    logger.error(f"Path segment {v} not mapped to a value.")  # noqa; E501
        if found_var_location:
            return global_, url
        if param.name not in self.path_vars:
            not_processed_path_vars = set(self.path_vars).difference(
                {v.key for v in url},
            )
            if len(not_processed_path_vars) > 1:
                err = f"Path variable ({param.name}) not in path and multiple path variables ({not_processed_path_vars}) to choose from."  # noqa; E501
                if self.mormo.strict:
                    raise ValueError(err)
                else:
                    logger.warning(err)
            elif len(not_processed_path_vars) == 1:
                first_var = not_processed_path_vars.pop()
                if first_var in mapped_value:
                    url.append(
                        Parameter(
                            key=first_var,
                            value=str(mapped_value[first_var]),
                        )
                    )
                else:
                    if mapped_value[param.name]:
                        # if self.strict:
                        #     raise ValueError("")
                        logger.warning(
                            f"Path variable doesn't exist in path ({param.name}), guessed to be path var ({self.path_vars[0]}) but not mapped to value"  # noqa; E501
                            " param is mapped to a value so using that value with the guessed name."  # noqa; E501
                        )
                        url.append(
                            Parameter(
                                key=self.path_vars[0],
                                value=str(mapped_value[param.name]),
                            )
                        )
                    else:
                        logger.error(f"Path variable doesn't exist in path ({param.name}).")  # noqa; E501
            else:
                logger.error(f"Path variable doesn't exist in path ({param.name}).")  # noqa; E501
        else:
            url.append(self.build_test_data_from_param(param, mapped_value))
        return global_, url

    def get_missing_path_variables(self, global_, url):
        global_ = global_.copy()
        url = url.copy()
        missing_variable = set(self.path_vars).difference(
            set([
                *[p.key for p in url if p],
                *[p.id for p in global_ if p],
            ]),
        )

        for path_var in missing_variable:
            mapped_value = self.get_mapped_value('path')
            if mapped_value.get(path_var):
                url.append(
                    Parameter(
                        key=path_var,
                        value=str(mapped_value[path_var]),
                    )
                )
            else:
                logger.warning(f"Path variable {path_var} missing variable and no mapped value!")  # noqa: E501
        return global_, url

    def get_request_body(self):
        header = []
        body_args = {}
        for param in self.params:
            param = self.mormo._resolve_object(param, new_cls=oapi.Parameter)
            if param.in_.value != "body":
                continue
            mapped_value = self.get_mapped_value('body')
            if not mapped_value:
                logger.warning(
                    f"Missing content for body parameter: {param.name}",
                )
                continue
            body_args['mode'] = 'raw'
            body_args['raw'] = json.dumps(dict(mapped_value))
            header.append(
                Parameter(
                    key='Content-Type',
                    value='application/json',
                ),
            )

        request_body = self.mormo._resolve_object(
            self.operation.requestBody,
            new_cls=oapi.RequestBody,
        )

        if (
            request_body and not body_args.get('mode') == 'raw'
            and self.config_test_data.get('requestBody')
        ):
            body_args['mode'] = 'raw'
            body_args['raw'] = json.dumps(
                self.config_test_data.get('requestBody'),
            )
        if body_args:
            return header, RequestBody(
                **body_args,
            )
        else:
            return header, None

    def build(self) -> PostmanVariables:
        global_, query, url, header, body = ([], [], [], [], [])
        for param in self.params:
            param = self.mormo._resolve_object(param, new_cls=oapi.Parameter)
            if not param.required:
                continue
            param_in = param.in_.value
            mapped_value = self.get_mapped_value(param_in)
            if param_in == 'body':
                continue
            test_data = Parameter(
                key=param.name,
                value=str(mapped_value[param.name]),
            )
            if param_in == 'path':
                n_global, n_url = self.get_path_param_variables(url, param)
                if global_ != n_global:
                    global_.extend(n_global)
                if url != n_url:
                    url.extend(n_url)
            elif param_in == 'query':
                query.append(test_data)
            elif param_in == 'header':
                header.append(test_data)
            elif param_in == 'cookie':
                raise ValueError(f"unhandled param location: {param_in}")
            else:
                raise ValueError(f"unknown param location: {param_in}")

        n_header, body = self.get_request_body()
        if header != n_header:
            header.extend(n_header)

        n_global, n_url = self.get_missing_path_variables(global_, url)
        if global_ != n_global:
            global_.extend(n_global)
        if url != n_url:
            url.extend(n_url)
        return PostmanVariables(global_, query, url, header, body)



from typing import Optional

from fastapi import FastAPI

from .convert import OpenAPIToPostman
from .schema import OpenAPISchemaToPostmanRequest, TestResult
from .schema.openapi_v3 import OpenAPISchemaV3, SaveDBResult
from .schema.postman_collection_v2 import (
    Collection,
    SaveDBResult as PostmanSaveDBResult,
)
from .util import DB, load_db, save_db
from . import logger, redis_handle, Settings

app = FastAPI(version='0.7.39')


@app.post("/schema", response_model=SaveDBResult)
def new_schema(o: OpenAPISchemaV3) -> SaveDBResult:
    """New Schema."""
    logger.debug("NEW SCHEMA")
    return save_db(o).to_dict()


@app.get('/schema/{id}', response_model=OpenAPISchemaV3)
def get_schema(id: str) -> OpenAPISchemaV3:
    return load_db(id)


@app.post("/schema/{id}/to_postman", response_model=PostmanSaveDBResult)
def schema_to_postman(
    id: str, o: Optional[OpenAPISchemaToPostmanRequest] = None,
) -> PostmanSaveDBResult:
    """Convert schema to Collection."""
    if o:
        kwargs = o.to_dict()
    else:
        kwargs = {}
    kwargs['schema_'] = load_db(id)
    o = OpenAPIToPostman(**kwargs).to_postman_collection_v2()
    return save_db(o).to_dict()


@app.get("/postman/{id}", response_model=Collection)
def get_postman(id: str) -> Collection:
    """Get Postman Collection by ID."""
    return load_db(id)


def run_postman_collection(
    collection, host: str, verbose: Optional[bool] = False,
):
    return TestResult(
        result=collection.run(host=host, verbose=verbose, json=True),
        code=0,
        message="Newman executed"
    )

# @app.post("/postman/{id}/test", response_model=TestResult)
# def run_postman_test_from_args(id: str, o: OpenAPISchemaToPostmanRequest) -> TestResult:
#     """Create a new test run from a postman collection."""
#     return run_postman_collection(OpenAPIToPostman(o), host=host, verbose=verbose)

@app.get("/postman/{id}/test", response_model=TestResult)
def run_postman_test(
    id: str, host: str, verbose: Optional[bool] = False,
) -> TestResult:
    """Create a new test run from a postman collection."""
    return run_postman_collection(load_db(id), host=host, verbose=verbose)


# @app.post('/test/run', response_model=SaveDBResult)
# def new_test_run(o: TestRun) -> SaveDBResult:
#     """Create a new test run."""
#     return save_db(o)


@app.post('/run/test/from_schema', response_model=TestResult)
def run_test_run_from_schema(o: OpenAPISchemaToPostmanRequest) -> TestResult:
    """Create a new test run from OpenAPI Schema."""
    from .convert import OpenAPIToPostman
    return run_postman_collection(
        OpenAPIToPostman(o).to_postman_collection_v2(),
        host=o.host,
        verbose=o.verbose,
    )

# @app.get('/test/run/{id}/fire', response_model=TestResult)
# def run_test_run(id):
#     load_db(id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mormo', 'mormo.schema']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.4.0,<0.5.0',
 'click>=7.0,<8.0',
 'fakeredis>=1.1.0,<2.0.0',
 'fastapi>=0.47.1,<0.48.0',
 'frozendict>=1.2,<2.0',
 'hypothesis-jsonschema>=0.10.2,<0.11.0',
 'jinja2>=2.10.3,<3.0.0',
 'pydantic>=1.4,<2.0',
 'python-dotenv>=0.10.5,<0.11.0',
 'pyyaml>=5.3,<6.0',
 'redis>=3.3.11,<4.0.0',
 'requests>=2.22.0,<3.0.0',
 'uvicorn>=0.11.2,<0.12.0']

entry_points = \
{'console_scripts': ['mormo = mormo:cli.cli']}

setup_kwargs = {
    'name': 'mormo',
    'version': '0.7.40',
    'description': 'Automated API Testing Framework for OpenAPI Schema.',
    'long_description': None,
    'author': 'Joey Stevens',
    'author_email': 'joey.stevens00@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
import json
from multiprocessing import Process
import sys
import tempfile
import time

import click
import requests
import uvicorn

from mormo.api import app
from mormo.convert import OpenAPIToPostman
from mormo.postman_test import run_newman


@click.group()
def cli():
    """OpenAPI to Postman Collection V2 Conversion."""
    pass


@cli.command()
def api():
    uvicorn.run(app, host="127.0.0.1", port=8001, log_level="info")


def generate_schema(infile, outfile, test_file, **kwargs):
    oas = OpenAPIToPostman(path=infile, test_data_file=test_file, **kwargs)
    postman_collection = oas.to_postman_collection_v2()
    with open(outfile, 'w') as f:
        json.dump(postman_collection.to_dict(), f)


@cli.command()
@click.option('-i', '--in', 'in_file', type=click.Path(),
              help='Path to the OpenAPI Schema to convert (YAML or JSON)', required=True)
@click.option('-t', '--test_file', 'test_file', type=click.Path(),
              help='Path to test config (YAML or JSON)')
@click.option('-o', '--out', 'out_file', type=click.Path(),
              help='The path to write the Postman Collection to.')
@click.option('--test', is_flag=True,
              help='Execute the generated schema with newman.')
@click.option('--test_mormo_api', is_flag=True,
              help='Spin up a mormo API with the binds in host.')
@click.option('--host', 'host', help='Target API')
@click.option('--verbose', is_flag=True,
              help='Verbose option in newman.')
def run(in_file, test_file, out_file, test, test_mormo_api, host, verbose):
    """Generate Postman Collections."""
    if not out_file:
        out_file = tempfile.mktemp()
    if test_mormo_api:
        test = True
        addr, port = host.split('/')[-1].split(':')
        port = port or 80
        proc = Process(
            target=uvicorn.run,
            args=(app,),
            kwargs={
                "host": addr,
                "port": int(port),
                "log_level": "info",
            },
            daemon=True,
        )
        proc.start()
        time.sleep(1)
        with open(in_file, 'w') as f:
            json.dump(requests.get(f'{host}/openapi.json').json(), f)
    generate_schema(in_file, out_file, test_file, host=host)
    if test:
        res = run_newman(out_file, host=host, verbose=verbose)
        if test_mormo_api:
            proc.terminate()
        sys.exit(res.code)


if __name__ == "__main__":
    cli()
